from time_tracker.db import TrackerDB, Task, WorkBlock
import time
import datetime
from pathlib import Path
from appdirs import AppDirs
import os


class TimeTracker:
    def __init__(self, session=None, epoch_time=None):
        """
        establish database connection, creates new one if database doesn't exist
        """
        if session:
            self.session = session
        else:
            dirs = AppDirs("tiempo_tracker", "dmikhr")
            db_dir = Path(dirs.user_data_dir)
            if not db_dir.is_dir():
                print(f'creating app directory: {str(db_dir)}')
                os.mkdir(str(db_dir))
            db_path = Path(f'{db_dir}/tracker.db')
            if not db_path.is_file():
                print(f'Database not found. Creating new one: {str(db_path)}')
                print("If you run app for the first time - it's normal behavior")
                TrackerDB(str(db_path)).create()
            self.session = TrackerDB(str(db_path)).connect()

        # can be used for testing purposes
        self.epoch_time = epoch_time

    def task_add(self, name, description=''):
        """add new task"""
        if not self._task_exists(name):
            self.session.add(
                            Task(name=name,
                                 description=description)
                            )
            self.session.commit()

            return f'Task {name} was added'
        else:
            return f'Task {name} already exists'

    # idea: replace if-check with decorator
    # @check(error_msg)
    def task_remove(self, name):
        """remove task"""
        if self._any_active_task():
            active_block = self._get_active_block()
            active_task = self._get_active_task(active_block)
            if active_task.name == name:
                return f"Can't delete task {name} while it's active"
        if self._task_exists(name):
            task_id = self._task_id(name)
            # to-do: wrap in transaction
            self.session.query(Task).filter(Task.id == task_id).delete()
            self.session.query(WorkBlock).filter(WorkBlock.task_id == task_id)\
                                         .delete()
            self.session.commit()
            # transaction end

            return f'Task {name} was removed'
        return f'Task {name} not found'

    def task_start(self, name):
        """
        start time tracking for a given :name: task
        """
        if not self._task_exists(name):
            return f'Task {name} not found. Maybe task was incorrectly typed?'
        # if there is an active task - stop it first
        # if active task is the same - show message
        if self._any_active_task():
            task_id = self._task_id(name)
            if self.session.query(WorkBlock).\
                    filter(WorkBlock.finish_time == None).\
                    one().task_id == task_id:
                return f'Task {name} is already in progress'
            self.task_finish()
        task_id = self._task_id(name)
        self.session.add(
                        WorkBlock(task_id=task_id,
                                  start_time=int(time.time()))
                        )
        self.session.commit()

        return f'Task {name} is in progress now...'

    def task_finish(self):
        """finish timke tracvking for an active task"""
        if not self._any_active_task():
            return 'No task is in progress. Start task first.'
        active_block = self.session.query(WorkBlock).\
                            filter(WorkBlock.finish_time == None).one()
        active_task = self.session.query(Task).\
                                   filter(Task.id == active_block.task_id)\
                                    .one()

        finish_time = int(time.time())

        active_block.finish_time = finish_time
        self.session.commit()
        self.session.flush()

        return (f'Task {active_task.name} was finished.'
                f'\nLast session time: {self._time_active_last(active_block.start_time, finish_time)}'
                f'\nTime in task today: {self._time_active_today(active_task.id)}'
                )

    def tasks_list(self):
        """
        show list of existing tasks
        if task tracking is active show '(in progress)' near task that is being tracked
        """
        active_task_id = -1
        tasks = self.session.query(Task).limit(-1).all()
        if self._any_active_task():
            active_task_id = self.session.query(WorkBlock).\
                                        filter(WorkBlock.finish_time == None).one().task_id
        tasks_output = []
        for task in tasks:
            tasks_output.append(f"{task.name}{' (in progress)' if task.id == active_task_id else ''}")
        return '\n'.join(tasks_output)

    def tasks_stats(self):
        """
        show how much time has been spent on each task during the day
        show only tasks that were active today
        time is calculated for today (with respect to day_starts in _time_active_today())
        time is presented as a timestamp 0:00 and as decimal number (e.g. 2.5)
        """
        today_active_tasks = {}
        # idea for future improvements: implement with join
        today_work_blocks_all = self.session.query(WorkBlock).\
                                             filter(WorkBlock.finish_time >= self._today_start_time()).all()
        # distinct on throws SADeprecationWarning, for reliability implement feature without it for now
        task_id_distinct = set(map(lambda item: item.task_id, today_work_blocks_all))
        for task_id in task_id_distinct:
            task_name = self.session.query(Task).filter(Task.id == task_id).one().name
            today_active_tasks[task_name] = self._time_active_today(task_id)
        # consider moving to separate method
        output_text = []
        for task, time_str in today_active_tasks.items():
            output_text.append(f'{task}\t{time_str}')
        return '\n'.join(output_text)

    def task_status(self):
        """
        show whether there is a task in progress
        if any task is active - show how much time it is active
        """
        # to-do: refactor task_finish and task_status
        if not self._any_active_task():
            return 'No task is active'
        active_block = self._get_active_block()
        active_task = self._get_active_task(active_block)
        return (f'Task in progress: {active_task.name}'
                f'\nCurrent session time: {self._time_active_last(active_block.start_time, int(time.time()))}'
                f'\nTime in task today: {self._time_active_today(active_task.id)}'
                )

    def _any_active_task(self):
        return len(self.session.query(WorkBlock).filter(WorkBlock.finish_time == None).all()) == 1

    def _get_active_block(self):
        return self.session.query(WorkBlock).filter(WorkBlock.finish_time == None).one()

    def _get_active_task(self, active_block):
        return self.session.query(Task).filter(Task.id == active_block.task_id).one()

    def _time_active_last(self, start_time, finish_time):
        time_delta = finish_time - start_time
        return str(datetime.timedelta(seconds=time_delta))

    def _time_active_today(self, task_id):
        """
        return amount of time spent on a given task today change when new day starts
        """
        today_work_blocks = self._work_blocks_today(task_id)
        today_work_time = 0
        for today_work_block in today_work_blocks:
            today_work_time += (today_work_block.finish_time - today_work_block.start_time)
        # in future development consider making _time_active_today return time in seconds and
        # add another method to format time for output to keep time calculation
        # and its representation separate
        time_formatted = self._round_time_and_fromat(str(datetime.timedelta(seconds=today_work_time)))
        time_decimal = self._time_to_decimal(time_formatted)
        return f'{time_formatted} ({time_decimal})'

    def _round_time_and_fromat(self, time_str):
        """
        removes seconds, round up if seconds >= 45 sec.
        '0:20:00' -> '0:20', '0:20:35' -> '0:20'
        assuming time blocks can't exceed 24 hrs - all stats currently works for 1 day data
        can be implemented with str formatting, but here for educational purposes implemented from scratch
        """
        hr, min, sec = time_str.split(':')
        # round up if seconds > 45
        if int(sec) > 45:
            if int(min) < 59:
                min = str(int(min) + 1)
                if int(min) < 10:
                    min = str(f'0{min}')
            elif int(min) == 59:
                min = '00'
                hr = str(int(hr) + 1)
        return f'{hr}:{min}'

    def _time_to_decimal(self, time_str):
        """
        convert timestamp like 0:00 to 0.0
        for long term time tracking in spreadsheets decimal time representation might be useful
        for things like time management, task prioritisation analysis, etc.
        """
        hr, min = time_str.split(':')
        return round((int(hr) * 60 + int(min)) / 60, 1)

    def _work_blocks_today(self, task_id):
        today_work_blocks = self.session.query(WorkBlock).\
                                         filter(WorkBlock.finish_time >= self._today_start_time()).\
                                         filter(WorkBlock.task_id == task_id).all()
        return today_work_blocks

    def _task_id(self, task_name):
        return self.session.query(Task).filter(Task.name == task_name).one().id

    def _task_exists(self, task_name):
        return len(self.session.query(Task).\
                                filter(Task.name == task_name).all()) == 1

    def _today_start_time(self):
        """
        in case you were working for example up to 1 A.M. it makes sence to track it as
        previous day  activity. By default set to 4 A.M.
        """
        day_starts = 4 # hrs
        seconds_in_hour = 60 * 60
        seconds_in_day = 24 * seconds_in_hour
        if self.epoch_time:
            epoch_time = self.epoch_time
        else:
            epoch_time = int(time.time())
        return epoch_time - (epoch_time % (seconds_in_day)) + day_starts * seconds_in_hour



import pytest
import string
from faker import Faker
import random
from time_tracker.time_tracker import TimeTracker
from time_tracker.db import Task, WorkBlock, TrackerDB
from sqlalchemy.orm import sessionmaker

TASKS_NUM = 3
WORK_BLOCKS_PER_TASK = 2
# time is always the same to make tests consistent
# set in epoch seconds
TIME_BASE = 1662374970
MINUTE = 60 # sec
HOUR = 60 * MINUTE

# generate random task name, currently not used to ensure tests consistency
def gen_task_name():
    return ''.join(list(map(
                            lambda _: random.choice(string.ascii_letters), 
                            range(random.randint(3,10))))).capitalize()


@pytest.fixture()
def test_db():
    fake = Faker()

    engine = TrackerDB(':memory:').create()
    Session = sessionmaker(bind=engine)
    session = Session()

    # gen names in list outside fixture so it can be used to check values
    for i in range(1, TASKS_NUM+1):
        session.add(
                    Task(name = f'Task {i}', 
                         description = fake.text(max_nb_chars=40))
                    )
        session.commit()

        for j in range (1, WORK_BLOCKS_PER_TASK+1):
            session.add(
                        WorkBlock(task_id = i,
                                  start_time = TIME_BASE + i * (j * 10 + j) * MINUTE,
                                  finish_time = TIME_BASE + i * (j * 20 + j) * MINUTE)
                        )
        session.commit()

    return session

def test_task_add(test_db):
    task_name = 'Test task'
    time_tracker = TimeTracker(test_db)
    time_tracker.task_add(task_name)

    assert test_db.query(Task).filter(Task.name == task_name).one().name == task_name
    assert len(test_db.query(Task).filter(Task.name == task_name).all()) == 1

    # can't be two tasks with same name
    time_tracker.task_add(task_name)
    assert len(test_db.query(Task).filter(Task.name == task_name).all()) == 1


def test_task_remove(test_db):
    # check if task is removed with all related work blocks
    task_to_delete = 'Task 1'
    task_id = test_db.query(Task).filter(Task.name == task_to_delete).one().id
    tasks_num_before = len(test_db.query(Task).limit(-1).all())

    work_blocks_before = len(test_db.query(WorkBlock).filter(WorkBlock.task_id == task_id).all())

    time_tracker = TimeTracker(test_db)
    time_tracker.task_remove(task_to_delete)
    tasks_num_after = len(test_db.query(Task).limit(-1).all())
    work_blocks_after = len(test_db.query(WorkBlock).filter(WorkBlock.task_id == task_id).all())

    assert tasks_num_after == tasks_num_before - 1
    assert work_blocks_before > 0
    assert work_blocks_after == 0

    # try to delete task that doesn't exist
    time_tracker.task_remove(task_to_delete)


def test_task_start(test_db):
    work_blocks_num_before = len(test_db.query(WorkBlock).limit(-1).all())
    time_tracker = TimeTracker(test_db)
    time_tracker.task_start('Task 1')
    work_blocks_num_after = len(test_db.query(WorkBlock).limit(-1).all())
    active_tasks_num = len(test_db.query(WorkBlock).\
                           filter(WorkBlock.finish_time == None).all())

    assert work_blocks_num_after == work_blocks_num_before + 1
    assert active_tasks_num == 1

    # can't start task that doesn't exist
    time_tracker.task_start('non_existent_task')
    work_blocks_num_after2 = len(test_db.query(WorkBlock).limit(-1).all())
    assert work_blocks_num_after == work_blocks_num_after2


def test_task_start_twice(test_db):
    # start same task twice
    time_tracker = TimeTracker(test_db)
    time_tracker.task_start('Task 1')
    active_task_id1 = test_db.query(WorkBlock).\
                              filter(WorkBlock.finish_time == None).one().task_id
    time_tracker.task_start('Task 1')
    active_task_id2 = test_db.query(WorkBlock).\
                              filter(WorkBlock.finish_time == None).one().task_id
    active_tasks_num = len(test_db.query(WorkBlock).\
                           filter(WorkBlock.finish_time == None).all())

    assert active_tasks_num == 1
    assert active_task_id1 == active_task_id2


def test_task_start_second(test_db):
    # start new task when another is in progress
    time_tracker = TimeTracker(test_db)
    time_tracker.task_start('Task 1')
    active_task_id1 = test_db.query(WorkBlock).\
                              filter(WorkBlock.finish_time == None).one().task_id
    time_tracker.task_start('Task 2')
    work_blocks_num_after = len(test_db.query(WorkBlock).limit(-1).all())
    active_task_id2 = test_db.query(WorkBlock).\
                              filter(WorkBlock.finish_time == None).one().task_id

    assert active_task_id1 ==  1
    assert active_task_id2 ==  2


def test_task_finish(test_db):
    time_tracker = TimeTracker(test_db)

    active_tasks_num_before1 = len(test_db.query(WorkBlock).\
                                 filter(WorkBlock.finish_time == None).all())
    time_tracker.task_start('Task 1')
    active_tasks_num_before = len(test_db.query(WorkBlock).\
                                 filter(WorkBlock.finish_time == None).all())

    time_tracker.task_finish()
    active_tasks_num_after = len(test_db.query(WorkBlock).\
                                 filter(WorkBlock.finish_time == None).all())

    assert active_tasks_num_before1 == 0
    assert active_tasks_num_before == 1
    assert active_tasks_num_after == 0

    # should handle correctly attempt to finish task twice
    time_tracker.task_finish()


def test_task_status(test_db):
    time_tracker = TimeTracker(test_db)
    response = time_tracker.task_status()
    assert response.find('Task in progress') == -1
    assert response.find('No task is active') > -1

    time_tracker.task_start('Task 1')
    response = time_tracker.task_status()
    assert response.find('Task in progress') > -1
    assert response.find('No task is active') == -1

def test_tasks_list(test_db):
    time_tracker = TimeTracker(test_db)
    tasks_list = time_tracker.tasks_list()
    assert set(tasks_list.split('\n')) == set([f'Task {i}' for i in range(1, TASKS_NUM+1)])

def test_tasks_stats(test_db):
    time_tracker = TimeTracker(test_db, epoch_time=TIME_BASE)
    # add task from previous day
    test_db.add(
                WorkBlock(task_id = 1,
                          start_time = TIME_BASE - 24 * HOUR,
                          finish_time = TIME_BASE - 24 * HOUR + 15 * MINUTE)
                )
    test_db.commit()

    time_tracker.tasks_stats()


import argparse
from time_tracker.time_tracker import TimeTracker


class Cli:
    def __init__(self):
        self.time_tracker = TimeTracker()
        self.methods_map = {'start': lambda task_name: self.time_tracker.task_start(task_name),
                            'add': lambda task_name: self.time_tracker.task_add(task_name),
                            'finish': lambda _: self.time_tracker.task_finish(),
                            'remove': lambda task_name: self.time_tracker.task_remove(task_name),
                            'list': lambda _: self.time_tracker.tasks_list(),
                            'stats': lambda _: self.time_tracker.tasks_stats()
                            }
        self.parser = argparse.ArgumentParser()
        self._setup_arguments()
        self.args = vars(self.parser.parse_args())
        self._run()

    def _run(self):
        if self._no_args():
            self.parser.print_help()
            print('\nNO PARAMETERS WERE SUPPLIED. SEE OPTIONS ABOVE')
            return
        for arg_name, method_name in self.methods_map.items():
            if self.args.get(arg_name):
                # print(f'calling with {arg_name}')
                response = method_name(self.args[arg_name])
                if response:
                    print(response)
                break

    def _setup_arguments(self):
        self.parser.add_argument('-s', '--start',
                            help='Start task by providing its name')
        # flag without value
        self.parser.add_argument('-f', '--finish', action='store_true',
                            help='Finish current task.\
                                 Task also can be finished\
                                 by starting a different task')
        self.parser.add_argument('-a', '--add',
                            help='Add new task')
        self.parser.add_argument('-r', '--remove',
                            help='Remove the task')
        self.parser.add_argument('-l', '--list', action='store_true',
                            help='List of all tasks')
        self.parser.add_argument('-st', '--stats', action='store_true',
                            help='Stats about tasks')

    def _no_args(self):
        return set(self.args.values()) == set([None, False])


from sqlalchemy import MetaData, Table, create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
Base = declarative_base()


class Task(Base):
   __tablename__ = 'tasks'
   
   id = Column(Integer, primary_key=True)
   name = Column(String)
   description = Column(String)
   children = relationship("WorkBlock", cascade="all,delete", backref="tasks")


class WorkBlock(Base):
   __tablename__ = 'work_blocks'
   
   id = Column(Integer, primary_key=True)
   task_id = Column(Integer, ForeignKey("tasks.id"))
   start_time = Column(Integer)
   finish_time = Column(Integer)

class TrackerDB:
   def __init__(self, db_name):
      self.db_name = db_name
      self.testing = False

   def connect(self):
      engine = self._engine()
      Session = sessionmaker(bind=engine)
      return Session()

   def create(self):
      engine = self._engine()
      self.meta_data = MetaData()
      self._create_tasks_table()
      self._create_work_block_table()
      self.meta_data.create_all(engine)
      return engine

   def _engine(self):
      if self.testing:
         return create_engine(f"sqlite+pysqlite:///{self.db_name}",
                              echo=True, future=True)
      else:
         return create_engine(f"sqlite+pysqlite:///{self.db_name}", future=True)

   def _create_tasks_table(self):
      return Table(
       'tasks', self.meta_data,
       Column('id', Integer, primary_key=True),
       Column('name', Integer),
       Column('description', Integer)
    )

   def _create_work_block_table(self):
      return Table(
       'work_blocks', self.meta_data,
       Column('id', Integer, primary_key=True),
       Column('task_id', Integer),
       Column('start_time', Integer),
       Column('finish_time', Integer)
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['time_tracker', 'time_tracker.tests']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.41,<2.0.0', 'appdirs>=1.4.4,<2.0.0']

entry_points = \
{'console_scripts': ['trt = time_tracker.__main__:main']}

setup_kwargs = {
    'name': 'tiempo-tracker',
    'version': '0.1.1',
    'description': 'Easy to use CLI time tracker written in Python',
    'long_description': ".. image:: https://api.codeclimate.com/v1/badges/bd649a97a34c7f8c4634/maintainability\n   :target: https://codeclimate.com/github/dmikhr/tiempo-tracker/maintainability\n   :alt: Maintainability\n\nIntroduction\n=========================\nTrack time you've spent on different tasks. Then see how much time you've spent on each. \n\nContinuation of `time_management <https://github.com/dmikhr/time_management>`_ script idea.\n\n\n**Development**\n=========================\nBuilt with Python, Poetry, SQLAlchemy for managing database, SQLite as database. \n\nTesting: Pytest\n\nRun package without building: ``poetry run python -m time_tracker [options]`` in the project directory\n\nExample: ``poetry run python -m time_tracker -a Task1``\n\nRun tests: ``poetry run pytest``\n\n\n**Installation**\n=========================\n**From source**\n\nrun ``make build`` and then ``make package-install`` in the package directory.\n\n**PIP**\n\nrun ``pip install tiempo-tracker``\n\nWhen you run app for the first time it will create a database. You'll see a message:\n\n``Database not found. Creating new one``\n\nAlso app will show path where database is stored.\n\n\n**Usage**\n=========================\ntiempo-tracker can be invoked by ``trt`` command in the terminal.\nAvailable options:\n::\n\n    -h, --help            show this help message and exit\n    -s START, --start START\n                            Start task by providing its name\n    -f, --finish          Finish current task. Task also can be finished \n                            by starting a different task\n    -a ADD, --add ADD     Add new task\n    -r REMOVE, --remove REMOVE\n                            Remove the task\n    -l, --list            List of all tasks\n    -st STATS, --stats STATS\n                            Stats about tasks\n\n\nStart with adding new task:\n\n``trt -a task1``\n\nThen track the task:\n\n``trt -s task1``\n\nTo finish the task either start a new one (add new task first if it doesn't exist yet):\n\n``trt -s task2``\n\nor finish task\n\n``trt -f``\n\nShow all added tasks:\n\n``trt -l``\n\nIf particular task is in progress there will be ``(in progress)`` status near this task. Example:\n::\n\n    Project_work\n    Exercising (in progress)\n    Flask_API_project\n    Some_task\n\nRemove task if it's no longer needed:\n\n``trt -r task1``\n\nSee how much time you've spent on each task during the day:\n\n``trt -st``\n\nTime will be presented in two formats: ``hr:min`` and as a decimal number. The latter is convenient for further storage in spreadsheets.",
    'author': 'Dmitrii Khramtsov',
    'author_email': 'dmitrypkh@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dmikhr/tiempo-tracker',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
